#This makes a 2 state FSA.  It does it without calling any of the 
#FSA test functions. State 0, turns on state 1, but both remain on.
#The first state also sends extra activation to state 1 so that it 
#goes as rapidly as possible.  The first is then shut off, and the second
#returns to 5ms firing.

from nealCoverClass import NealCoverFunctions
import pyNN.spiNNaker as sim

import cPickle as pickle
import numpy as np

from stateMachineClass import FSAHelperFunctions

#---Below here are functions to test the system---
RUN_DURATION = 300
#initialize the simulator. 
def init():
    sim.setup(timestep=neal.DELAY,min_delay=neal.DELAY,
                    max_delay=neal.DELAY, debug=0)

def printPklSpikes(fileName):
    fileHandle = open(fileName)
    neoObj = pickle.load(fileHandle)
    segments = neoObj.segments
    segment = segments[0]
    spikeTrains = segment.spiketrains
    neurons = len(spikeTrains)
    for neuronNum in range (0,neurons):
        if (len(spikeTrains[neuronNum])>0):
            spikes = spikeTrains[neuronNum]
            for spike in range (0,len(spikes)):
                print neuronNum, spikes[spike]
    fileHandle.close()
    

def createTwoInputs():
    inputSpikeTimes0 = [10.0]
    inputSpikeTimes1 = [150.0]
    spikeArray0 = {'spike_times': [inputSpikeTimes0]}
    spikeGen0=sim.Population(1,sim.SpikeSourceArray,spikeArray0,
                                   label='inputSpikes_0')
    spikeArray1 = {'spike_times': [inputSpikeTimes1]}
    spikeGen1=sim.Population(1, sim.SpikeSourceArray, spikeArray1,
                                   label='inputSpikes_1')

    return [spikeGen0,spikeGen1]


def createNeurons():
    numNeurons = fsa.CA_SIZE * 2
        
    cells=sim.Population(numNeurons,sim.IF_cond_exp,fsa.CELL_PARAMS)
    return cells

def setupRecording(cells):
    if (neal.simulator == 'nest'):
        cells.record({'spikes','v'})
    elif  (neal.simulator == 'spinnaker'):
        cells.record()
        cells.record_v()

def printResults(simCells):
    if  (neal.simulator == 'nest'):
        simCells.printSpikes('temp.pkl')
        #printPklSpikes('temp.pkl')
    elif (neal.simulator == 'spinnaker'):
        simCells.printSpikes('temp.sp')
        simCells.print_v('temp.volts')

##--main---
simName = "spinnaker"
spinnVersion = 7
neal = NealCoverFunctions(simName,sim,spinnVersion)
fsa = FSAHelperFunctions(simName,sim,neal,spinnVersion)
init()
spikeGenerators = createTwoInputs()
firstSpikeGenerator =  spikeGenerators[0]
secondSpikeGenerator =  spikeGenerators[1]

stateCells = createNeurons()

setupRecording(stateCells)

#Build the FSA
fsa.turnOnStateFromSpikeSource(firstSpikeGenerator,stateCells,0)
fsa.makeCA(stateCells,0)
fsa.makeCA(stateCells,1)
fsa.stateTurnsOnState(stateCells,0,stateCells,1)
fsa.stateStimulatesState(stateCells,0,stateCells,1,1.0)
fsa.turnOffStateFromSpikeSource(secondSpikeGenerator,stateCells,0)

neal.nealApplyProjections()
sim.run(RUN_DURATION)

printResults(stateCells)



