#This makes 1 state, and uses a spike source to turn it on, stimulate it,
#inhibit it, and then shut it off.

from nealCoverClass import NealCoverFunctions
import pyNN.spiNNaker as sim

import cPickle as pickle
import numpy as np

from stateMachineClass import FSAHelperFunctions

#---Below here are functions to test the system---
RUN_DURATION = 300
#initialize the simulator. 
def init():
    sim.setup(timestep=neal.DELAY,min_delay=neal.DELAY,
                    max_delay=neal.DELAY, debug=0)

def printPklSpikes(fileName):
    fileHandle = open(fileName)
    neoObj = pickle.load(fileHandle)
    segments = neoObj.segments
    segment = segments[0]
    spikeTrains = segment.spiketrains
    neurons = len(spikeTrains)
    for neuronNum in range (0,neurons):
        if (len(spikeTrains[neuronNum])>0):
            spikes = spikeTrains[neuronNum]
            for spike in range (0,len(spikes)):
                print neuronNum, spikes[spike]
    fileHandle.close()
    

def createFourInputs():
    inputSpikeTimes0 = [10.0]
    spikeArray0 = {'spike_times': [inputSpikeTimes0]}
    spikeGen0=sim.Population(1,sim.SpikeSourceArray,spikeArray0,
                                   label='inputSpikes_0')

    inputSpikeTimes1 = [100.0]
    spikeArray1 = {'spike_times': [inputSpikeTimes1]}
    spikeGen1=sim.Population(1,sim.SpikeSourceArray,spikeArray1,
                                   label='inputSpikes_1')

    inputSpikeTimes2 = [150.0]
    spikeArray2 = {'spike_times': [inputSpikeTimes2]}
    spikeGen2=sim.Population(1,sim.SpikeSourceArray,spikeArray2,
                                   label='inputSpikes_2')

    inputSpikeTimes3 = [200.0]
    spikeArray3 = {'spike_times': [inputSpikeTimes3]}
    spikeGen3=sim.Population(1,sim.SpikeSourceArray,spikeArray3,
                                   label='inputSpikes_3')

    return [spikeGen0,spikeGen1,spikeGen2,spikeGen3]


def createNeurons():
    numNeurons = fsa.CA_SIZE * 2
    cells=sim.Population(numNeurons,sim.IF_cond_exp,fsa.CELL_PARAMS)
    return cells

def setupRecording(cells):
    if  ((neal.simulator == 'nest') or 
         ((neal.simulator == 'spinnaker') and (neal.spinnVersion == 8))):
        cells.record({'spikes','v'})
    elif ((neal.simulator == 'spinnaker') and (neal.spinnVersion == 7)):
        cells.record()
        cells.record_v()

def printResults(simCells):
    if  ((neal.simulator == 'nest') or 
         ((neal.simulator == 'spinnaker') and (neal.spinnVersion == 8))):
        simCells.printSpikes('temp.pkl')
        #printPklSpikes('temp.pkl')
    elif ((neal.simulator == 'spinnaker') and (neal.spinnVersion == 7)):
        simCells.printSpikes('temp.sp')

##--main---
simName = "spinnaker"
spinnVersion = 8
neal = NealCoverFunctions(simName,sim,spinnVersion)
fsa = FSAHelperFunctions(simName,sim,neal,spinnVersion)
init()
spikeGenerators =  createFourInputs()

stateCells = createNeurons()
setupRecording(stateCells)

#make a state
fsa.makeCA(stateCells,0)

#play with the state using spike sources
fsa.turnOnStateFromSpikeSource(spikeGenerators[0],stateCells,0)
fsa.stimulateStateFromSpikeSource(spikeGenerators[1],stateCells,0,0.1)
fsa.inhibitStateFromSpikeSource(spikeGenerators[2],stateCells,0,-0.2)
fsa.turnOffStateFromSpikeSource(spikeGenerators[3],stateCells,0)

neal.nealApplyProjections()
sim.run(RUN_DURATION)

printResults(stateCells)
